import ipaddress
from pathlib import Path
from typing import Optional, Union
from urllib.error import HTTPError
from urllib.request import urlopen

Numeric = Union[int, float]


def ask_host(prompt: str, default: str = "127.0.0.1", allow_hostname=False) -> str:
    while True:
        user_reply = input(f"{prompt} (default: {default}): ")
        if user_reply == "":
            user_reply = default
        try:
            if allow_hostname:
                url = user_reply
                if not (user_reply.startswith("http://") or user_reply.startswith("https://")):
                    url = f"http://{user_reply}"
                try:
                    urlopen(url)
                    break
                except HTTPError:
                    print("Please input correct URL.")
            ipaddress.ip_address(user_reply)
            break
        except ValueError:
            print("Please input correct host.")
    return user_reply


def convert_str_into_numeric(user_reply: str) -> Numeric:
    if user_reply.isdigit():
        return int(user_reply)
    return float(user_reply)


def ask_number(prompt: str, default: Numeric, min_value: Numeric, max_value: Numeric) -> Numeric:
    while True:
        user_reply = input(f"{prompt} (default: {default}): ")
        try:
            if user_reply == "":
                return default
            if (
                user_reply.isdigit()
                and min_value <= convert_str_into_numeric(user_reply) <= max_value
            ):
                user_reply_numeric = convert_str_into_numeric(user_reply)
                break
        except ValueError:
            print(f"Please input correct number between {min_value}~{max_value}.")
    return user_reply_numeric


def ask_string(prompt: str, default: str = "", use_default: bool = True) -> str:
    while True:
        if use_default:
            user_reply = input(f'{prompt} (default: "{default}"): ')
            if user_reply == "":
                return default
            return user_reply
        else:
            user_reply = input(f"{prompt} (if you don't want, just leave empty): ")
            return user_reply


def ask_string_in_array(prompt: str, choices: list, default: str) -> Optional[str]:
    if default and default not in choices:
        print("Default value should be in choices args.")
        return None
    if "" in choices:
        choices.remove("")

    if default:
        question = (
            f"{prompt} (choices: {'/'.join(choices)}, "
            f"if left empty, this will use default value: {default}): "
        )
    else:
        question = (
            f"{prompt} (choices: {'/'.join(choices)}, if left empty, this will remove this key): "
        )

    while True:
        user_reply = input(question)
        if user_reply == "":
            if default:
                user_reply = default
            else:
                return None
            break
        elif user_reply.lower() in choices:
            break
        else:
            print(f"Please answer in {'/'.join(choices)}.")
    return user_reply


def ask_path(prompt: str, is_file=True, is_directory=True) -> Path:
    if not (is_file or is_directory):
        print("One of args(is_file/is_directory) has True value.")
    while True:
        user_reply = input(f"{prompt}: ")
        path = Path(user_reply)
        if is_file and path.is_file():
            break
        if is_directory and path.is_dir():
            break

        if is_file and is_directory:
            print("Please answer a correct file/directory path.")
        elif is_file:
            print("Please answer a correct file path.")
        elif is_directory:
            print("Please answer a correct directory path.")
    return path


def ask_yn(prompt: str = "Are you sure?", default: str = "y") -> bool:
    if default == "y":
        choices = "Y/n"
    elif default == "n":
        choices = "y/N"
    else:
        raise ValueError("default must be given either 'y' or 'n'.")
    while True:
        user_reply = input("{0} [{1}] ".format(prompt, choices)).lower()
        if user_reply == "":
            user_reply = default
        if user_reply in ("y", "yes", "n", "no"):
            break
        else:
            print("Please answer in y/yes/n/no.")
    if user_reply[:1].lower() == "y":
        return True
    return False


def ask_tf(prompt: str = "Are you sure?", default: str = "true") -> bool:
    if default == "t":
        choices = "T/f"
    elif default == "f":
        choices = "t/F"
    else:
        raise ValueError("default must be given either 'true' or 'n'.")
    while True:
        user_reply = input(f"{prompt} [{choices}] ").lower()
        if user_reply == "":
            user_reply = default
        if user_reply in ("t", "true", "f", "false"):
            break
        else:
            print("Please answer in t/true/f/false.")
    if user_reply[:1].lower() == "t":
        return True
    return False



import shutil

from .loader import load_entry_points

main = load_entry_points()
# main object is called by the console script.

if __name__ == "__main__":
    # Execute right away if the module is directly called from CLI.
    main(max_content_width=shutil.get_terminal_size().columns - 2)



import click

from .extensions import ExtendedCommandGroup
from .types import CliContextInfo


@click.group(
    cls=ExtendedCommandGroup,
    context_settings={
        "help_option_names": ["-h", "--help"],
    },
)
@click.option(
    "--skip-sslcert-validation",
    help="<Client option> Skip SSL certificate validation for all API requests.",
    is_flag=True,
)
@click.option(
    "--output",
    type=click.Choice(["json", "console"]),
    default="console",
    help="<Client option> Set the output style of the command results.",
)
@click.pass_context
def main(ctx: click.Context, **kwargs) -> None:
    """Unified Command Line Interface for Backend.ai"""
    ctx.obj = CliContextInfo(info=kwargs)



import os
import signal
import sys

import click
from click.exceptions import Abort, ClickException

from .types import ExitCode


class InterruptAwareCommandMixin(click.BaseCommand):
    """
    Replace the main() method to support proper exit-codes
    for interruptions on Windows and POSIX platforms.
    Using this, interrupting the command will let the shell
    know that the execution is also interrupted instead of
    continuing the shell/batch script.
    """

    def main(self, *args, **kwargs):
        try:
            _interrupted = False
            kwargs.pop("standalone_mode", None)
            kwargs.pop("prog_name", None)
            super().main(
                *args,
                standalone_mode=False,
                prog_name="backend.ai",
                **kwargs,
            )
        except KeyboardInterrupt:
            # For interruptions outside the Click's exception handling block.
            print("Interrupted!", end="", file=sys.stderr)
            sys.stderr.flush()
            _interrupted = True
        except Abort as e:
            # Click wraps unhandled KeyboardInterrupt with a plain
            # sys.exit(1) call and prints "Aborted!" message
            # (which would look non-sense to users).
            # This is *NOT* what we want.
            # Instead of relying on Click, mark the _interrupted
            # flag to perform our own exit routines.
            if isinstance(e.__context__, KeyboardInterrupt):
                print("Interrupted!", end="", file=sys.stderr)
                sys.stderr.flush()
                _interrupted = True
            else:
                print("Aborted!", end="", file=sys.stderr)
                sys.stderr.flush()
                sys.exit(ExitCode.FAILURE)
        except ClickException as e:
            e.show()
            sys.exit(e.exit_code)
        finally:
            if _interrupted:
                # Override the exit code when it's interrupted,
                # referring https://github.com/python/cpython/pull/11862
                if sys.platform.startswith("win"):
                    # Use STATUS_CONTROL_C_EXIT to notify cmd.exe
                    # for interrupted exit
                    sys.exit(-1073741510)
                else:
                    # Use the default signal handler to set the exit
                    # code properly for interruption.
                    signal.signal(signal.SIGINT, signal.SIG_DFL)
                    os.kill(os.getpid(), signal.SIGINT)


class AliasGroupMixin(click.Group):
    """
    Enable command aliases.

    ref) https://github.com/click-contrib/click-aliases
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._commands = {}
        self._aliases = {}

    def command(self, *args, **kwargs):
        aliases = kwargs.pop("aliases", [])
        decorator = super().command(*args, **kwargs)
        if not aliases:
            return decorator

        def _decorator(f):
            cmd = decorator(f)
            if aliases:
                self._commands[cmd.name] = aliases
                for alias in aliases:
                    self._aliases[alias] = cmd.name
            return cmd

        return _decorator

    def group(self, *args, **kwargs):
        aliases = kwargs.pop("aliases", [])
        # keep the same class type
        kwargs["cls"] = type(self)
        decorator = super().group(*args, **kwargs)
        if not aliases:
            return decorator

        def _decorator(f):
            cmd = decorator(f)
            if aliases:
                self._commands[cmd.name] = aliases
                for alias in aliases:
                    self._aliases[alias] = cmd.name
            return cmd

        return _decorator

    def get_command(self, ctx, cmd_name):
        if cmd_name in self._aliases:
            cmd_name = self._aliases[cmd_name]
        command = super().get_command(ctx, cmd_name)
        if command:
            return command

    def format_commands(self, ctx, formatter):
        commands = []
        for subcommand in self.list_commands(ctx):
            cmd = self.get_command(ctx, subcommand)
            # What is this, the tool lied about a command. Ignore it
            if cmd is None:
                continue
            if cmd.hidden:
                continue
            if subcommand in self._commands:
                aliases = ",".join(sorted(self._commands[subcommand]))
                subcommand = "{0} ({1})".format(subcommand, aliases)
            commands.append((subcommand, cmd))

        # allow for 3 times the default spacing
        if len(commands):
            limit = formatter.width - 6 - max(len(cmd[0]) for cmd in commands)
            rows = []
            for subcommand, cmd in commands:
                help = cmd.get_short_help_str(limit)
                rows.append((subcommand, help))
            if rows:
                with formatter.section("Commands"):
                    formatter.write_dl(rows)


class ExtendedCommandGroup(InterruptAwareCommandMixin, AliasGroupMixin, click.Group):
    pass



import enum
from typing import Dict

import attr


@attr.define(slots=True)
class CliContextInfo:
    info: Dict = attr.field()


class ExitCode(enum.Enum):
    OK = 0
    FAILURE = 1
    TIMEOUT = 2
    INVALID_ARGUMENTS = 3



import click  # noqa: E402

from ai.backend.plugin.entrypoint import scan_entrypoints

from .main import main  # noqa: E402


def load_entry_points() -> click.Group:
    entry_prefix = "backendai_cli_v10"
    for entrypoint in scan_entrypoints(entry_prefix):
        if entrypoint.name == "_":
            cmd_group: click.Group = entrypoint.load()
            for name, cmd in cmd_group.commands.items():
                main.add_command(cmd, name=name)
        else:
            prefix, _, subprefix = entrypoint.name.partition(".")
            if not subprefix:
                subcmd = entrypoint.load()
                main.add_command(subcmd, name=prefix)
            else:
                subcmd = entrypoint.load()
                main.commands[prefix].add_command(subcmd, name=subprefix)  # type: ignore
    return main




# DO NOT EDIT THIS FILE -- AUTOGENERATED BY PANTS
# Target: src/ai/backend/cli:dist

from setuptools import setup

setup(**{
    'author': 'Lablup Inc. and contributors',
    'classifiers': [
        'Intended Audience :: Developers',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Environment :: No Input/Output (Daemon)',
        'Topic :: Scientific/Engineering',
        'Topic :: Software Development',
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: MIT License',
    ],
    'description': 'Backend.AI Command Line Interface Helper',
    'entry_points': {
        'console_scripts': [
            'backend.ai = ai.backend.cli.__main__:main',
        ],
    },
    'install_requires': (
        'attrs>=20.3',
        'backend.ai-plugin==22.09.0b1',
        'click>=7.1.2',
        'types-click',
    ),
    'license': 'MIT',
    'long_description': """# backend.ai-cli

Unified command-line interface for Backend.AI


## How to adopt in CLI-enabled Backend.AI packages

An example `setup.cfg` in Backend.AI Manager:
```
[options.entry_points]
backendai_cli_v10 =
    mgr = ai.backend.manager.cli.__main__:main
    mgr.start-server = ai.backend.gateway.server:main
```

Define your package entry points that returns a Click command group using a
prefix, and add additional entry points that returns a Click command using a
prefix followed by a dot and sub-command name for shortcut access, under the
`backendai_cli_v10` entry point group.

Then add `backend.ai-cli` to the `install_requires` list.

You can do the same in `setup.py` as well.
""",
    'long_description_content_type': 'text/markdown',
    'name': 'backend.ai-cli',
    'namespace_packages': (
    ),
    'package_data': {
        'ai.backend.cli': (
            'VERSION',
            'py.typed',
        ),
    },
    'packages': (
        'ai.backend.cli',
    ),
    'project_urls': {
        'Documentation': 'https://docs.backend.ai/',
        'Source': 'https://github.com/lablup/backend.ai',
    },
    'python_requires': '==3.10.5',
    'url': 'https://github.com/lablup/backend.ai',
    'version': '22.09.0b1',
    'zip_safe': False,
})
